/*****************************************************************************
 * This software is Copyright (C) 1998 by Rafal Wierzbicki. I accept no
 * responsibility for anything this software may or may not do to your system
 * - you use it completely at your own risk. This software comes under the GPL
 * licence. See the COPYING file for full legal details.
 *****************************************************************************/

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include "../../configure.h"
#include "../../include/module.h"
#include "../../include/aftersteplib.h"
#include "../../include/afterstep.h"

#include "Gnome.h"

#define LOG(s) fprintf (stderr, "%s\n", s)

typedef struct node
{
  long id;
  long flags;
  int workspace;
  struct node *previous;
  struct node *next;
}list_item;

typedef struct 
{
  list_item *first, *last;
}s_list;

s_list *window_list;
static int current_desk = 0;

list_item *
s_list_insert (s_list *list, long id)
{
  list_item *item = NULL;

  item = (list_item *) malloc (sizeof (list_item));

  if (!item)
    return NULL;

  if (list->first == NULL)
    {      
      list->first = item;
      list->last = item;
    }
  else
    {
      list->last->next = item;
    }
  item->next = NULL;
  list->last = item;
  item->id = id;
  return item;
}

/* blah blah blah i was tired when i wrote this one, ok? */
int s_list_remove_by_data (s_list *list, long id)
{
  list_item *tmp1, *tmp2;

  tmp1 = list->first;
  tmp2 = NULL;

  if (tmp1->id == id)
    {
      list->first = tmp1->next;
      free(tmp1);
      return 1;
    }
  
  while (tmp1)
    {
      
      if (tmp1->id == id)
	{
	  if (tmp1->id == list->last->id)
	    {
	      list->last = tmp2;
	      tmp2->next = NULL;
	    }
	  else
	    {
	      if (tmp2 != NULL)
		tmp2->next = tmp1->next;
	      else
		list->first->next = tmp1->next;
	    }	      
	  free (tmp1);
	  return 1;
	}
      tmp2 = tmp1;
      tmp1 = tmp1->next;
    }
  return 0;
}

list_item * s_list_find_by_data (s_list *list, long id)
{
  list_item *tmp;

  tmp = list->first;

  if (list->first == NULL && list->last == NULL)
    return NULL;

  while (tmp)
    {
      if (tmp->id == id)
	return tmp;
      tmp=tmp->next;
    }
  return NULL;
}

s_list *s_list_new ()
{
  s_list *list;

  list = (s_list *) malloc (sizeof (s_list));

  if (!list)
    return NULL;

  list->first = list->last = NULL;
  return list;
}
int s_list_free (s_list *list)
{
  list_item *tmp1, *tmp2;

  tmp1=list->first;
  while (tmp1)
    {
      tmp2 = tmp1->next;
      free(tmp1);
      tmp1=tmp2;
    }
  free (list);
  return 0;
}

int s_list_count (s_list *list)
{
  list_item *tmp;
  int count = 0;

  tmp = list->first;

  while (tmp)
    {
      count++;
      tmp = tmp->next;
    }
  return count;
}

static void 
gnome_compliance_init ()
{
  Atom supported_list[10];
  int count;

  root_win = RootWindow (dpy, screen);

  /* supporting WM check */
  _XA_WIN_SUPPORTING_WM_CHECK = XInternAtom (dpy, "_WIN_SUPPORTING_WM_CHECK", 
					     False);
  _XA_WIN_PROTOCOLS = XInternAtom(dpy, "_WIN_PROTOCOLS", False);
  _XA_WIN_LAYER = XInternAtom(dpy, "_WIN_LAYER", False);      
  _XA_WIN_STATE = XInternAtom(dpy, "_WIN_STATE", False);       
  _XA_WIN_HINTS = XInternAtom(dpy, "_WIN_HINTS", False);
  _XA_WIN_APP_STATE = XInternAtom(dpy, "_WIN_APP_STATE", False);
  _XA_WIN_EXPANDED_SIZE = XInternAtom(dpy, "_WIN_EXPANDED_SIZE", False);
  _XA_WIN_ICONS = XInternAtom(dpy, "_WIN_ICONS", False);
  _XA_WIN_WORKSPACE = XInternAtom(dpy, "_WIN_WORKSPACE", False);
  _XA_WIN_WORKSPACE_COUNT = XInternAtom(dpy, "_WIN_WORKSPACE_COUNT", False);
  _XA_WIN_WORKSPACE_NAMES = XInternAtom(dpy, "_WIN_WORKSPACE_NAMES", False);
  _XA_WIN_CLIENT_LIST = XInternAtom(dpy, "_WIN_CLIENT_LIST", False);
  _XA_WIN_DESKTOP_BUTTON_PROXY = XInternAtom(dpy, "_WIN_DESKTOP_BUTTON_PROXY", 
					     False);

  /* create the GNOME window */
  gnome_win = XCreateSimpleWindow (dpy, root_win, 0, 0, 5, 5, 0, 0, 0);

  /* supported WM check */
  XChangeProperty (dpy, root_win, _XA_WIN_SUPPORTING_WM_CHECK,
		   XA_CARDINAL, 32, PropModeReplace, (unsigned char *)&gnome_win, 1);
  XChangeProperty (dpy, gnome_win , _XA_WIN_SUPPORTING_WM_CHECK,
		   XA_CARDINAL, 32, PropModeReplace, (unsigned char *) &gnome_win, 1);
 
  /* desktop button proxy */
  XChangeProperty (dpy, root_win, _XA_WIN_DESKTOP_BUTTON_PROXY, XA_CARDINAL,
		   32, PropModeReplace, (unsigned char *) &gnome_win, 1);
  XChangeProperty (dpy, gnome_win, _XA_WIN_DESKTOP_BUTTON_PROXY, XA_CARDINAL,
		   32, PropModeReplace, (unsigned char *) &gnome_win, 1);
  
  /* supported protocols, some are not so supported so they really shouldn't be here */
  count = 0;

  supported_list[count++] = _XA_WIN_LAYER;
  supported_list[count++] = _XA_WIN_STATE;
  supported_list[count++] = _XA_WIN_HINTS;
  supported_list[count++] = _XA_WIN_APP_STATE;
  supported_list[count++] = _XA_WIN_EXPANDED_SIZE;
  supported_list[count++] = _XA_WIN_ICONS;
  supported_list[count++] = _XA_WIN_WORKSPACE;
  supported_list[count++] = _XA_WIN_WORKSPACE_COUNT;
  supported_list[count++] = _XA_WIN_WORKSPACE_NAMES;
  supported_list[count++] = _XA_WIN_CLIENT_LIST;

  XChangeProperty(dpy, root_win, _XA_WIN_PROTOCOLS, XA_ATOM, 32, PropModeReplace,
		  (unsigned char *)supported_list, 10);

}

static void
gnome_set_current_workspace (int current_desk)
{

  XChangeProperty (dpy, root_win, _XA_WIN_WORKSPACE, XA_CARDINAL, 32, 
		   PropModeReplace,  (unsigned char *) &current_desk, 1);
}

static void 
gnome_set_workspace_count (int workspaces)
{

  XChangeProperty (dpy, root_win, _XA_WIN_WORKSPACE_COUNT, XA_CARDINAL, 32, 
		   PropModeReplace, (unsigned char *) &workspaces, 1);

}

void
gnome_set_workspace_names (int count, char **names)
{
  XTextProperty text;

  if (XStringListToTextProperty (names, count, &text))
      {
	XSetTextProperty (dpy, root_win, &text, _XA_WIN_WORKSPACE_NAMES);
	XFree(text.value);
      }
}

void
gnome_set_client_list (s_list *list)
{
  Window *windows;
  int windows_count, count;
  list_item *tmp;
  
  windows_count = s_list_count (list);

  if (windows_count == 0)
    return;
  windows = malloc (sizeof (Window) * windows_count);
  if (!windows)
    {
      fprintf (stderr, "gnome_set_client_list: malloc failed\n");
      return;
    }

  tmp = list->first;
  count = 0;

  while (tmp)
    {
      windows[count++] = tmp->id;
      tmp=tmp->next;
    }
  XChangeProperty (dpy, root_win, _XA_WIN_CLIENT_LIST, XA_CARDINAL, 32,
		   PropModeReplace, (unsigned char *) windows, windows_count);
}

void
gnome_set_win_hints (list_item *aswin)
{
  long flags=0;

  if (aswin->flags & STICKY)
    flags |= WIN_STATE_STICKY;
  
  if (aswin->flags & ICONIFIED)
    flags |= WIN_STATE_MINIMIZED;

  if (aswin->flags & SHADED)
    flags |= WIN_STATE_SHADED;

  if (aswin->workspace != current_desk)
    flags |= WIN_STATE_HID_WORKSPACE;

  XChangeProperty (dpy, (Window) aswin->id,  _XA_WIN_WORKSPACE, XA_CARDINAL,
		   32, PropModeReplace, (unsigned char *)&aswin->workspace, 1);

  XChangeProperty (dpy, (Window) aswin->id, _XA_WIN_STATE, XA_CARDINAL, 
		   32, PropModeReplace, (unsigned char*)&flags, 1);
}

void
gnome_handle_win_layer (XClientMessageEvent *event)
{
  LOG ("_XA_WIN_LAYER not implemented yet");
}

void
gnome_handle_win_state (XClientMessageEvent *event)
{
  long new_members , change_mask;
  list_item *aswin;

  aswin = s_list_find_by_data (window_list, (long) event->window);
  change_mask = event->data.l[0];
  new_members = event->data.l[1];
  
  if (new_members & WIN_STATE_STICKY)
    SendText (fd, "Stick 1\n", aswin->id);
  else if (change_mask & WIN_STATE_STICKY)
    SendText (fd, "Stick 0\n", aswin->id);

  if (new_members & WIN_STATE_MAXIMIZED_VERT)
    SendText (fd, "Maximize 0 100\n", aswin->id);
  
    LOG("WIN_STATE_MAXIMIZED_VERT not implemented yet");
  if (new_members & WIN_STATE_MAXIMIZED_HORIZ)
    LOG("WIN_STATE_MAXIMIZED_VERT not implemented yet");

  if (new_members & WIN_STATE_SHADED)
    SendText (fd, "Shade 1\n", aswin->id);
  else if (change_mask & WIN_STATE_SHADED)
    SendText (fd, "Shade 0\n", aswin->id);

  if (new_members & WIN_STATE_MINIMIZED)
    SendText (fd, "Iconify 1\n", aswin->id);
  else if (change_mask & WIN_STATE_MINIMIZED)
    SendText (fd, "Iconify 0\n", aswin->id);
}

void
DeadPipe (int sig)
{

  switch (sig)
    {
    case SIGSEGV:
      fprintf (stderr, "segfault\n");
      exit (-1);
    case SIGINT:
      LOG ("User abort, exiting");
      exit (-1);
    case SIGPIPE:
      fprintf (stderr, "pipe\n");
    default:
      break;
    }
#ifdef DEBUG_ALLOCS
 print_unfreed_mem();
#endif 
  exit(0);
}

/* set_as_mask:
 * set the mask on the pipe
 */

void set_as_mask (long unsigned mask)
{
  char set_mask_mesg[255];

  sprintf (set_mask_mesg, "SET_MASK %lu\n", mask);
  SendText (fd, set_mask_mesg, 0);
}

/* send_as_mesg:
 * send multiple  message to the pipe, have to be seperated by a ','
 */
 
void send_as_mesg (char *message, unsigned long id)
{
  char *delim;
  char tmp[255];
  char token[50];

  sprintf (tmp, message);

  while  ((delim = index (tmp, ',')) != NULL )
    {
      snprintf (token, strlen(tmp) - strlen(delim) +1, tmp);
      sprintf (tmp, delim + 1);
      SendText (fd, token, id);
      delim = index (tmp, ',');	    
    }
  SendText (fd, tmp, id);
}


void
process_message (unsigned long type, int elements, unsigned long *body)
{
  int status = 0;

  switch (type)
    {
    case M_CONFIGURE_WINDOW:
      status = list_configure (body);
      break;
    case M_ADD_WINDOW:
      status = list_add_window (body);
      break;
    case M_DESTROY_WINDOW:
      status = list_destroy_window (body);
      break;
    case M_ICONIFY:
      status = list_iconify (body);
      break;
    case M_DEICONIFY:
      status = list_deiconify (body);
      break;
    case M_WINDOW_NAME:
      status = list_window_name (body);
      break;
    case M_ICON_NAME:
      status = list_icon_name (body);
      break;
    case M_END_WINDOWLIST:
      status = list_end ();
      break;
    case M_NEW_PAGE:
      status = list_page_change (body);
      break;
    case M_NEW_DESK:
      status = list_desk_change (body);
      break;
    default:
      break;
    }
}

/* list_configure:
 * pipe configure events
 */
    
int
list_configure (unsigned long *body)
{
  list_item *item;

  if (body[8] & WINDOWLISTSKIP)
    return 0;

  if (body[0] == gnome_win)
    {
      return 0;
    }

  if ((item = s_list_find_by_data (window_list, body[0])) != NULL)
      {
	item->flags = body[8];
	item->workspace = body[7];
	gnome_set_win_hints (item);
	return 1;
      }

  if ((item=s_list_insert (window_list, body[0])) == NULL)
    {    
      fprintf (stderr, "failed in list_configure\n");
      return 0;
    }

  item->flags = body[8];
  item->workspace = body[7];
  gnome_set_win_hints (item);
  gnome_set_client_list (window_list);

  return 1;
}

/* list_add_window:
 * pipe add window events
 */

int
list_add_window (unsigned long *body)
{
  list_item *item;

  if (s_list_find_by_data (window_list, body[0])) 
      return 0;

  if (body[8] & WINDOWLISTSKIP)
    return 0;

  if (body[0] == gnome_win)
    return 0;

  item = s_list_insert (window_list, body[0]);
  item->flags = body[8];
  item->workspace = body[7];
  gnome_set_win_hints (item);
  gnome_set_client_list (window_list);
  return 1;
}

/* list_window_name:
 * pipe chage of window name events
 */

int
list_window_name (unsigned long *body)
{
  return 0;
}

/* list_icon_name:
 * pipe change of icon name events
 */

int
list_icon_name (unsigned long *body)
{
  return 0;
}

/* list_destroy_window:
 * pipe destroyed windows event
 */

int
list_destroy_window (unsigned long *body)
{
  if (s_list_find_by_data (window_list, body[0]))
    {
      s_list_remove_by_data (window_list, body[0]);
      gnome_set_client_list (window_list);
      return 1;
    }
  return 0;
}

/* list_end
 * end of window list from pipe
 */
 
int
list_end (void)
{
  /* there is no way to get a number of workspaces and their names so, here, 4 workspaces :) */
  /* this will have to be done through a config file, maybe Pagers cfg file */
  char *work_spaces[4];
  int i;

  fprintf (stderr, "%i windows counted\n", s_list_count (window_list));
  gnome_set_current_workspace (current_desk);
  gnome_set_workspace_count (4);

  for (i=0;i<4;i++)
    {
      work_spaces[i] = safemalloc (20);
      sprintf (work_spaces[i], "Desktop %i", i);
    }
  gnome_set_workspace_names (4, work_spaces);
  gnome_set_client_list (window_list);
  for (i=0;i <4; i++)
    {
      free (work_spaces[i]);
    }
  return 0;
}

/* list_deiconify:
 * window deiconified from the pipe
 */

int
list_deiconify (unsigned long *body)
{
  return 0;
}

/* list_iconify:
 * window iconified from the pipe
 */

int
list_iconify (unsigned long *body)
{
  return 0;
}

int
list_page_change (unsigned long *body)
{
  return 0;
  current_desk = (int) body[2];

  if (body[2] != 10000)
    {
      gnome_set_current_workspace (current_desk);
    }
  return 0;
}

int
list_desk_change (unsigned long *body)
{
  if (body[0] != 10000)
    {
      current_desk = (unsigned int) body[0];
      fprintf (stderr, "new desk: %lu\n", body[0]);      
      gnome_set_current_workspace (current_desk);
    }
  return 0;
}
static int
error_handler (Display *disp, XErrorEvent *event)
{
  fprintf (stderr, "%s: internal error, error code %d, request code %d, minor code %d\n  . Please email rafal@mcss.mcmaster.ca with te above information\n", MyName, event->error_code, event->request_code, event->minor_code);
  return 0;
}

static void
event_handler ()
{
  XEvent Event;
  char tmp[255];

  while (XPending (dpy))
    {
      XNextEvent (dpy, &Event);
      switch (Event.type)
	{
	case PropertyNotify:
	  break;
	case ClientMessage:
	  /* change workspace request */
	  if (Event.xclient.message_type == _XA_WIN_WORKSPACE)
	    {
	      int desk;
	      desk = Event.xclient.data.l[0];
	      fprintf (stderr, "desk change requested - %d\n", desk);
	      sprintf (tmp, "Desk 10000 %d\n", desk);
	      SendInfo (fd, tmp, None);
	    }
	  /* change layer request */
	  else if (Event.xclient.message_type == _XA_WIN_LAYER)
	    {
	      LOG("_XA_WIN_LAYER request");
	      gnome_handle_win_layer (&Event.xclient);
	    }
	  /* state requests */
	  else if (Event.xclient.message_type == _XA_WIN_STATE)
	    {
	      LOG("_XA_WIN_STATE request");
	      gnome_handle_win_state (&Event.xclient);
	    }
	  break;
	default:
	  break;
	}      
    }
}
/******************************************************************************
  EndLessLoop -  Read and redraw until we get killed, blocking when can't read
******************************************************************************/
static void
EndLessLoop ()
{
  fd_set readset;
  struct timeval tv;

  while (1)
    {
      FD_ZERO (&readset);
      FD_SET (fd[1], &readset);
      FD_SET (x_fd, &readset);
      tv.tv_sec = 0;
      tv.tv_usec = 0;
#ifdef __hpux
      if (!select (fd_width, (int *) &readset, NULL, NULL, &tv))
	{
	  XPending (dpy);
	  FD_ZERO (&readset);
	  FD_SET (fd[1], &readset);
	  FD_SET (x_fd, &readset);
	  select (fd_width, (int *) &readset, NULL, NULL, NULL);
	}
#else
      if (!select (fd_width, &readset, NULL, NULL, &tv))
	{
	  XPending (dpy);
	  FD_ZERO (&readset);
	  FD_SET (fd[1], &readset);
	  FD_SET (x_fd, &readset);
	  select (fd_width, &readset, NULL, NULL, NULL);
	}
#endif
      if (FD_ISSET (x_fd, &readset))
	event_handler ();
      if (!FD_ISSET (fd[1], &readset))
	continue;
      read_as_socket ();
    }
}

static void
read_as_socket ()
{
  unsigned long header[3], *body;
  if (ReadASPacket (fd[1], header, &body) > 0)
    {
      process_message (header[1], header[2], body);
      free (body);
    }
}

void version(void)
{
  printf("%s version %s\n", MyName, VERSION);
  exit(0);
}

void usage(void)
{
  printf("Usage:\n"
	 "%s [--version] [--help]\n", MyName);
  exit(0);
}

int main (int argc, char **argv)
{
  char *temp;
  int i;
  char *global_config_file = NULL;

  /* Save our program name - for error messages */
  temp = strrchr (argv[0], '/');
  MyName = temp ? temp + 1 : argv[0];

  for (i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
      if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
	usage ();
      else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--version"))
	version ();
      else if (!strcmp(argv[i], "-w") || !strcmp(argv[i], "--window"))
	i++;
      else if (!strcmp(argv[i], "-c") || !strcmp(argv[i], "--context"))
	i++;
      else if (!strcmp(argv[i], "-f") && i + 1 < argc)
	global_config_file = argv[++i];
    }

  /* Dead pipe == dead AfterStep */
  signal (SIGPIPE, DeadPipe);
  signal (SIGSEGV, DeadPipe);
  signal (SIGINT, DeadPipe);

  if ((dpy = XOpenDisplay ("")) == NULL)
    {
      fprintf (stderr, "%s: couldn't open display %s\n",
	       MyName, XDisplayName (""));
      exit (1);
    }
  screen = DefaultScreen (dpy);

  /* connect to AfterStep */
  temp = module_get_socket_property(RootWindow(dpy, screen));
  fd[0] = fd[1] = module_connect (temp);
  XFree (temp);
  if (fd[0] < 0)
    {
      fprintf (stderr, "%s: unable to establish connection to AfterStep\n", MyName);
      exit (1);
    }
  temp = safemalloc (9 + strlen(MyName) + 1);
  sprintf (temp, "SET_NAME %s", MyName);
  SendInfo (fd, temp, None);
  free (temp);
  
  x_fd = XConnectionNumber (dpy);
  fd_width = GetFdWidth ();

  XSetErrorHandler (error_handler);

  gnome_compliance_init ();

  window_list = s_list_new ();

  set_as_mask ((long unsigned) mask_reg);
  SendText (fd, "Send_WindowList\n", 0);

  XSelectInput (dpy, root_win, PropertyChangeMask | SubstructureNotifyMask);
  XSelectInput (dpy, gnome_win, PropertyChangeMask);

  EndLessLoop ();
  return 0;
}
