#ifndef PARSER_HEADER_FILE_INCLUDED
#define PARSER_HEADER_FILE_INCLUDED

/*#define DEBUG_PARSER*/

typedef struct term_definition {
    /* the following needs to be defined by module */
#define TF_DONT_REMOVE_COMMENTS (1<<0) /* indicates that comments sould not be */ 
                                       /* removed from the end of this term */
#define TF_SPECIAL_PROCESSING 	(1<<1) /* supplied special processing procedure */
                                       /* needs to be called for this one */
#define TF_NO_MYNAME_PREPENDING	(1<<2) /* has effect only when writing this back to */
                                       /* config file */
#define TF_DONT_SPLIT           (1<<3) /* if it is desired to preserv data intact */
                                       /* vs. splitting it into space separated words */
    unsigned long flags ; 	/* combination of any of above values */
    char* keyword ;         
    unsigned int keyword_len ;
#define TT_INTEGER  1
#define TT_COLOR    2
#define TT_FILENAME 3
#define TT_PATHNAME 4
#define TT_GEOMETRY 5
#define TT_TEXT     6
#define TT_CUSTOM   7           /* modules can define custom types starting */
                                /* with this one */
    int type ;                  /* term's type */

#define TT_INTEGER_CUSTOM_ID  0 /* that is done for some global config id's */
#define TT_COLOR_CUSTOM_ID    0 /* all module's custom ID's should start from */
#define TT_FILENAME_CUSTOM_ID 0 /* here */
#define TT_PATHNAME_CUSTOM_ID 0
#define TT_GEOMETRY_CUSTOM_ID 0
#define TT_TEXT_CUSTOM_ID     0
#define TT_CUSTOM_CUSTOM_ID   0
    int id ;                    /* term's id whithin the type */
    /* the following will be initialized automatically */
    struct term_definition* brother ; /* next term with the same hash key in hash table */
} TermDef ;

typedef struct syntax_definition {
    char terminator ;          /* character, terminating this term */
    char file_terminator ;     /* character, terminating configuration */
    TermDef* terms ;        /* array of Term definitions */ 
} SyntaxDef ;

typedef struct freestorage_elem {
    TermDef* term ;           
    char**   argv ;           /* space separated words from the source data will
                                 be placed here unless DONT_SPLIT_WORDS defined 
				 for the term*/
    int      argc ;           /* number of words */
    struct freestorage_elem* next ;
} FreeStorageElem ;

#define TERM_HASH_SIZE 		61
#define MAX_CONFIG_LINE 	MAXLINELENGTH

typedef unsigned short int HashValue ;

typedef struct config_definition {
    SyntaxDef* syntax ;
    TermDef**   term_hash ;
    HashValue   term_hash_size ;

    int (*special)(struct config_definition* conf_def, FreeStorageElem** storage);
    char* realfilename ;
    FILE* fd ;
    int bNeedToCloseFile ;
    /* allocated to store lines read from the file */
    char* buffer ;
    /* this is the current parsing information */
    char* tline ;     
    TermDef* current_term ;
    char* current_data;
    int current_data_len ;
    
    char* cursor ;

} ConfigDef ;

ConfigDef* InitConfigReader( SyntaxDef* syntax, const char* filename, FILE* fd, HashValue hash_size, int (*special)(struct config_definition* conf_def, FreeStorageElem** storage) );
int ParseConfig( ConfigDef* config, FreeStorageElem** storage );
void DestroyConfigReader(ConfigDef* config);
void DestroyFreeStorage(FreeStorageElem** storage );

#ifdef DEBUG_PARSER
void PrintConfigReader(ConfigDef* config);
void PrintFreeStorage(FreeStorageElem* storage);
#else
#define PrintConfigReader(a)
#define PrintFreeStorage(b)
#endif

#define COMMENTS_CHAR '#'
#define MYNAME_CHAR   '*'

#endif


